/**
 ** p3dDisplayTaskCodels.c
 **
 ** Codels appeles par la tache p3dDisplayTask
 **
 ** Auteur: Thierry Peynot
 ** Date: 
 **
 **/

#ifndef UNIX
#include <vxWorks.h>
#include <timers.h>
#else
#include <portLib.h>
#endif

#ifndef HAS_POSIX_CLOCK
#include <sys/time.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "libp3d.h"
#include "server/p3dHeader.h"

/* Other libs */
#ifdef GDHE
#include <gdhe/GDHE.h>
#endif

/* To be able to display positions */
#include <t3d/t3d.h>

/* Constants */
#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

/* Global variables */



#define DEBUG 0

/* for temporary display functions, not used if Interface is ON */
#define TEMP 1


/*------------------------------------------------------------------------
 *
 * p3dDisplayTaskInit  --  Codel d'initialisation (SDI/f, ...) de la tache DisplayTask
 *
 * Description: 
 *
 * Description: initializes the display params with default values (executed at
 * the module initialisation)
 * 
 * Retourne:    OK ou ERROR
 */

STATUS p3dDisplayTaskInit(int *bilan)
{
  /* Display initialisation and display params */
  strncpy (SDI_F->displayParams.serverName, P3D_DEFAULT_DISPLAY_SERVER,
	   P3D_MAX_STRLEN);
  SDI_F->displayParams.arcLength = 3.0;
  SDI_F->displayParams.numNodes = 10;
  
  SDI_F->displayState.gdheConnection = P3D_GDHE_NOT_CONNECTED;

  /* Init current robot position in SDI (before reading PoM Poster) */
  SDI_F->p3dPosRobotToOrigin.thetaRobot = 0.0;
  SDI_F->p3dPosRobotToOrigin.pitchRobot = 0.0;
  SDI_F->p3dPosRobotToOrigin.rollRobot = 0.0;
  SDI_F->p3dPosRobotToOrigin.xRobot = 0.0;
  SDI_F->p3dPosRobotToOrigin.yRobot = 0.0;
  SDI_F->p3dPosRobotToOrigin.zRobot = 0.0;

  return OK;
}


/*------------------------------------------------------------------------
 *
 * p3dDisplayTaskEnd  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS p3dDisplayTaskEnd(void)
{
#ifdef GDHE
  if (SDI_F->displayState.gdheConnection == P3D_GDHE_CONNECTED)
    disconnect ();
#endif

  return OK;
}


#ifdef GDHE
/*------------------------------------------------------------------------
 * InitDisplay
 *
 * Description: Initializes a display. Opens a connection with gdhe, 
 * displays a first time
 *
 * Bilans:      OK 
 *              S_p3dDisplayTask_CONNECTION_FAILED
 *		S_p3dDisplayTask_BAD_DISPLAY_PARAMS
 *              S_p3dDisplayTask_POM_POSTER_NOT_FOUND
 * 
 */

/* p3dInitDisplayExec  -  codel EXEC de InitDisplay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dInitDisplayExec (P3D_DISPLAY_PARAMS *displayParams, int *bilan)
{
  float  cylHeight=2.0;  /* height of the GDHE cylinder illustrating a waypoint */
  float  cylDiameter=0.5; /* Diameter of that cylinder */

  char	 gdheCmdBuf[1024];

  static POSTER_ID pomPosterId;
  POM_POS pos;
  double  xRobot, yRobot, zRobot, thetaRobot;


  /* Check some params values ? */
  
  /*--------- Get connection with Gdhe --------------*/
  if (SDI_F->displayState.gdheConnection == P3D_GDHE_CONNECTED)
    disconnect ();

  if (get_connection(displayParams->serverName) == FALSE) {
    *bilan = S_p3d_CONNECTION_FAILED;
    return ETHER;
  }
  SDI_F->displayState.gdheConnection = P3D_GDHE_CONNECTED;


  /*----- Prepare Gdhe cylinders to illustrate wayPoints -----*/
  /* The first waypoint is in red, the others in black */
  /* First waypoint */
  p3dDispCreateFirstWayPoint(cylDiameter,cylHeight);

  /* Other waypoints */
  p3dDispCreateWayPoints(P3D_MAX_WAYPOINTS-1, cylDiameter, cylHeight);

  printf("p3dInitDisplay: %d GDHE waypoints cylinders prepared\n", P3D_MAX_WAYPOINTS);


  /*----- Get the Robot Position ----------*/
  if (!pomPosterId) {
     if (posterFind("pomPos", &pomPosterId) == ERROR) {
       fprintf(stderr, "p3dInitDisplayExec: cannot find pom poster\n");
       *bilan = S_p3d_POM_POSTER_NOT_FOUND;
    }
  }
   if (p3dPOM_POSPosterRead(pomPosterId, &(pos)) == ERROR) {
    fprintf(stderr, "p3dInitDisplayExec: unable to read pomPos poster\n");
    *bilan = S_p3d_POM_POSTER_NOT_FOUND;
  }
  xRobot = pos.mainToOrigin.euler.x;
  yRobot = pos.mainToOrigin.euler.y;
  zRobot = pos.mainToOrigin.euler.z;
  thetaRobot = pos.mainToOrigin.euler.yaw;


  /*-- Prepare arc to illustrate v,w ref computed by p3d ---*/
  sprintf (gdheCmdBuf, 
	   "set robots(arc) {polyline 2 0 0 0 2 0 0}" );
  eval_expression (gdheCmdBuf);
  
#if DEBUG
  printf("p3dDisplay: %s\n", gdheCmdBuf);
#endif

  /*-- pos --*/
  p3dDispSetPosArc(thetaRobot, xRobot, yRobot, zRobot);


  /*-- Display the Global Frame in red, pos. of center: Origin --*/
  eval_expression("set robots(frame) {color 255 0 0 ; repere }");
  eval_expression("set pos(frame) {0 0 0 0 0 0.2}");


  /*-- Prepare mini dalas to illustrate predicted config. */
  p3dDispCreateMiniDalas(P3D_DEFAULT_MAX_NUMNODES, displayParams->scale);

  printf("p3dInitDisplay: %d mini Dala prepared, scale = %f\n", P3D_DEFAULT_MAX_NUMNODES, displayParams->scale);

  return ETHER;
}


/* p3dStopDisplayExec  -  codel EXEC de StopDisplay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dStopDisplayExec (int *bilan)
{
  
  /* Delete Gdhe objects ? */

  /* Stops the Connection with gdhe if still exists*/
  if (SDI_F->displayState.gdheConnection == P3D_GDHE_CONNECTED)
    disconnect ();

  SDI_F->displayState.gdheConnection = P3D_GDHE_NOT_CONNECTED;

  return ETHER;
}



/*------------------------------------------------------------------------
 * UpdateDisplay
 *
 * Description: updates a display
 *
 * Bilans:      OK 
 *              S_p3dDisplayTask_NOT_CONNECTED
 *
 */
/* p3dUpdateDisplayExec  -  codel EXEC de UpdateDisplay
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dUpdateDisplayExec (int *bilan)
{
  float  x,y;   /* for wayPoints coordinates */
  double  xRobot, yRobot, zRobot, thetaRobot, pitchRobot, rollRobot;
  
  int   numNodes=SDI_F->displayParams.numNodes;  /* Number of nodes for the polyline */
  double   length=SDI_F->displayParams.arcLength;      /* Length of the arc drawn */
  int  i;
  

  /* Check the connection with gdhe */
  if (SDI_F->displayState.gdheConnection == P3D_GDHE_NOT_CONNECTED) {
    *bilan = S_p3d_NOT_CONNECTED;
    return ETHER;
  }

  
  /*----- Display cylinders on the locations of wayPoints available ----*/
#if DEBUG
  printf("p3dUpdateDisplay: %d waypoints in the list\n", SDI_F->wpList.number);
#endif

  for(i = 0; i < SDI_F->wpList.number; i++)
    {
      /* Read the coordinates of this wayPoint */
      x = SDI_F->wpList.wps[i].x;
      y = SDI_F->wpList.wps[i].y;

      /* send the command to gdhe to draw the cylinder*/
      p3dDispSetPosWaypoint(i,x,y);
    }

  /* TODO: the wayPoints no longer used should be deleted */
  

  /*------- Display an arc illustrating the ref speeds v,w in poster -----*/
  
  /*----- Get the Robot Position ----------*/
  xRobot = SDI_F->p3dPosRobotToOrigin.xRobot;
  yRobot = SDI_F->p3dPosRobotToOrigin.yRobot;
  zRobot = SDI_F->p3dPosRobotToOrigin.zRobot;
  thetaRobot = SDI_F->p3dPosRobotToOrigin.thetaRobot;
  pitchRobot = SDI_F->p3dPosRobotToOrigin.pitchRobot;
  rollRobot  = SDI_F->p3dPosRobotToOrigin.rollRobot;

  /*------ Gdhe Arc to illustrate p3d's v,w ref ----------------*/
  p3dDispDrawSpeedRefArc(SDI_F->speedRef.v, SDI_F->speedRef.w, numNodes, length);


  /*---- The arc starts at Robot's position ----*/
  p3dDispSetPosArc(thetaRobot, xRobot, yRobot, zRobot);
  

#if TEMP
  /* This should be done only if the Interface is not used */
  /* Display the robot at its position used to make planning */
  p3dDispSetPosDala(thetaRobot, pitchRobot, rollRobot, xRobot, yRobot, zRobot);
  /*-- Roll is set to 0.0 for the moment --*/
#endif
  
  return ETHER;
}



/* Convert a predicted configuration into a T3D (type=T3DEuler) */
/* returns that T3D */
T3D*
arcNodeConfigGenToT3D(P3DNodeGen*    node)    /* configuration */
{
  T3D*  nodeToRobot;

  /* Read the position for the current arc node and put it in the T3D "nodeToRobot" */
  /*  (this is a relative position wrt the robot) */
  nodeToRobot = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  nodeToRobot->euler.euler[0] = node->theta;
  nodeToRobot->euler.euler[1] = 0.0;
  nodeToRobot->euler.euler[2] = 0.0;  /* roll */
  nodeToRobot->euler.euler[3] = node->x;
  nodeToRobot->euler.euler[4] = node->y;
  nodeToRobot->euler.euler[5] = 0.0;

  return(nodeToRobot);
}		   



/* p3dDisplayPredictedConfigExec  -  codel EXEC de DisplayPredictedConfig
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT p3dDisplayPredictedConfigExec (P3D_DISPLAYCONFIG_PARAMS *displayConfigParams, int *bilan)
{
  T3D*   nodeToRobot;
  T3D*   robotToOrigin;

  P3DArcGen bestCurrentArc = SDI_F->path.arcs[0];
  int   nodeNb;
#if DEBUG
  int   k;
#endif

  /* Check params */
  if( (displayConfigParams->firstNodeNb >= bestCurrentArc.n) ||
      (displayConfigParams->lastNodeNb >= bestCurrentArc.n)  ||
      (displayConfigParams->firstNodeNb > displayConfigParams->lastNodeNb) ||
      displayConfigParams->interval <= 0 )
    {
      printf("p3dDisplay: There are only %d in the arc\n", bestCurrentArc.n);
      *bilan = S_p3d_BAD_DISPLAY_PARAMS;
      return ETHER;
    }

  robotToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  /* Need only the 2D position.  TODO: Check that !! */
  robotToOrigin->euler.euler[0] = SDI_F->p3dPosRobotToOrigin.thetaRobot;
  robotToOrigin->euler.euler[1] = 0.0;
  robotToOrigin->euler.euler[2] = 0.0;
  robotToOrigin->euler.euler[3] = SDI_F->p3dPosRobotToOrigin.xRobot;
  robotToOrigin->euler.euler[4] = SDI_F->p3dPosRobotToOrigin.yRobot;
  robotToOrigin->euler.euler[5] = SDI_F->p3dPosRobotToOrigin.zRobot;

#if DEBUG
  printf("p3dDisplay: robotToOrigin = ");
  for(k=0; k<6; k++)
    printf("%f ", robotToOrigin->euler.euler[k]);
  printf("\n");
#endif

  for(nodeNb = displayConfigParams->firstNodeNb; 
      nodeNb <= displayConfigParams->lastNodeNb;
      nodeNb += displayConfigParams->interval )
    {
      nodeToRobot = arcNodeConfigGenToT3D(&(bestCurrentArc.nodes[nodeNb]));
      
#if DEBUG
      printf("p3dDisplay: nodeToRobot[%d] = ", nodeNb);
      for(k=0; k<6; k++)
	printf("%f ", nodeToRobot->euler.euler[k]);
      printf("\n");
#endif

      p3dDisplayOnePredictedConfig(nodeToRobot,
				   bestCurrentArc.nodes[nodeNb].conf.pitch, 
				   bestCurrentArc.nodes[nodeNb].conf.roll, 
				   bestCurrentArc.nodes[nodeNb].conf.z, 
				   robotToOrigin,
				   nodeNb, SDI_F->displayParams.scale);
    }
  
  return ETHER;
}
#endif


